unit IEDriveInfo;

{==================================================================

 Component TDriveInfo  /  Version 1.1  / 5.1999
 ===============================================


    Description:
    ============
    Central drive management class. Provides information about all
    installed drives, for example wether drive is valid, disk is inserted
    displayname or volume size.

    Author:
    =======
    (c) Ingo Eckel 1999
    Sodener Weg 38
    65812 Bad Soden
    Germany

    For detailed documentation and history see the documentation in TDriveInfo.htm.

{==================================================================}

{Required compiler options:}
{$A+,B-,X+,H+,P+}

interface

uses
  Windows,
  SysUtils,
  Classes,
  ComCtrls,
  ShellApi,
  Math,
  CommCtrl;


      {Flags used by TDriveInfo.ReadDriveStatus and TDriveView.RefreshRootNodes:}
const dsValid             = 0;  {checks only whether drive is still valid}
      dsImageIndex        = 1;  {Fetch imageindex, if not allready fetched}
      dsSize              = 2;  {Fetch disk size and serialnumber}
      dsDisplayName       = 4;  {Fetch drives displayname}
      dsAll               = dsImageIndex  Or
                            dsSize        Or
                            dsDisplayName;

      FirstDrive          = 'A';
      FirstFixedDrive     = 'C';
      LastDrive           = 'Z';


Type  TDrive    = Char;

      TDriveInfoRec = Record
        Init        : Boolean;     {Drivestatus was updated once}
        Valid       : Boolean;     {Drivestatus is valid}
        DriveReady  : Boolean;     {Drive is ready}
        DriveType   : Integer;     {DRIVE_REMOVABLE, DRIVE_FIXED, DRIVE_CDROM, DRIVE_RAMDISK, DRIVE_REMOTE}
        DisplayName : String;      {Windows displayname}
        Prettyname  : String;      {Prettyfied displayname}
        LongPrettyName : String;   {UNC-Network name on Network drives or same as PrettyName}
        DriveSerial : DWORD;       {Serial number of the drive}
        Size        : Int64;       {Drivesize}
        ImageIndex  : Integer;     {Drive imageIndex}
        FileSystemName  : String;  {Filesystemname as returned by GetVolumeInformation}
        FileSystemFlags : DWORD;   {Filesystem flags as returned by GetVolumeInformation}
      End;


{---------------------------------------------------------------}
  TDriveInfo = Class(TObject)
{---------------------------------------------------------------}
  Private
{---------------------------------------------------------------}
    FData : Array[FirstDrive..LastDrive] Of TDriveInfoRec;
    Function GetData(Drive : TDrive) : TDriveInfoRec;
{---------------------------------------------------------------}
  Public
{---------------------------------------------------------------}
    Property Data[Drive : TDrive]          : TDriveInfoRec Read GetData; default;

    Function GetImageIndex(Drive : TDrive) : Integer;
    Function GetDisplayName(Drive : TDrive) : String;
    Function GetPrettyName(Drive : TDrive) : String;
    Function GetLongPrettyName(Drive : TDrive) : String;
    Function ReadDriveStatus(Drive : TDrive; Flags : Integer) : Boolean;
    Constructor Create;
    Destructor Destroy;  Override;
  End;
{---------------------------------------------------------------}

Function GetShellFileName (Const Name : String ) : String;
Function GetNetWorkName   (Drive : Char) : String;


{Central drive information object instance of TDriveInfo}
Var DriveInfo : TDriveInfo;

ResourceString
  ErrorInvalidDrive = '%s is a invalid drive letter.';


{---------------------------------------------------------------}
implementation
{---------------------------------------------------------------}


// ===========================================================
// Class TDriveInfo:
// ===========================================================

Constructor TDriveInfo.Create;
Var Drive : TDrive;

Begin
 Inherited Create;
 For Drive := FirstDrive To LastDrive Do
 With FData[Drive] Do
 Begin
   Init        := False;
   DriveType   := Windows.GetDriveType(PChar(Drive + ':\'));
   Valid       := DriveType in [DRIVE_REMOVABLE, DRIVE_FIXED, DRIVE_CDROM, DRIVE_RAMDISK, DRIVE_REMOTE];
   Init        := False;
   DriveReady  := False;
   DisplayName := '';
   PrettyName  := '';
   LongPrettyName := '';
   FileSystemName := '';
   DriveSerial := 0;
   Size        := -1;
   ImageIndex  := 0;
 End;
End; {TDriveInfo.Create}


Destructor TDriveInfo.Destroy;
Var Drive : TDrive;
Begin
  For Drive := FirstDrive To LastDrive Do
  With FData[Drive] Do
  Begin
    SetLength(DisplayName, 0);
    SetLength(PrettyName, 0);
    SetLength(LongPrettyName, 0);
    SetLength(FileSystemName, 0);
  End;
  Inherited Destroy;
End; {TDriveInfo.Destroy}


Function TDriveInfo.GetImageIndex(Drive : TDrive) : Integer;
Begin
  if (Drive < FirstDrive) Or (Drive > LastDrive) then
    Raise EConvertError.Create(Format(ErrorInvalidDrive, [Drive]));

  Result := 0;

  IF FData[Drive].Valid Then
  Begin
    IF (FData[Drive].ImageIndex = 0) Then
    ReadDriveStatus(Drive, dsImageIndex);
    Result := FData[Drive].ImageIndex;
  End;
End; {TDriveInfo.GetImageIndex}


Function TDriveInfo.GetDisplayName(Drive : TDrive) : String;
Begin
  if (Drive < FirstDrive) Or (Drive > LastDrive) then
    Raise EConvertError.Create(Format(ErrorInvalidDrive, [Drive]));

  Result := Drive + ':';

  IF FData[Drive].Valid Then
  Begin
    IF (Length(FData[Drive].DisplayName) = 0) Then
    ReadDriveStatus(Drive, dsDisplayName);
    Result := FData[Drive].DisplayName;
  End;
End; {TDriveInfo.GetDisplayname}


Function TDriveInfo.GetPrettyName(Drive : TDrive) : String;
Begin
  if (Drive < FirstDrive) Or (Drive > LastDrive) then
    Raise EConvertError.Create(Format(ErrorInvalidDrive, [Drive]));

  Result := Drive + ':';

  IF FData[Drive].Valid Then
  Begin
    IF (Length(FData[Drive].PrettyName) = 0) Then
    ReadDriveStatus(Drive, dsDisplayName);
    Result := FData[Drive].PrettyName;
  End;
End; {TDriveInfo.GetPrettyName}


Function TDriveInfo.GetLongPrettyName(Drive : TDrive) : String;
Begin
  if (Drive < FirstDrive) Or (Drive > LastDrive) then
    Raise EConvertError.Create(Format(ErrorInvalidDrive, [Drive]));

  Result := Drive + ':';

  IF FData[Drive].Valid Then
  Begin
    IF (Length(FData[Drive].PrettyName) = 0) Then
    ReadDriveStatus(Drive, dsDisplayName);
    Result := FData[Drive].LongPrettyName;
  End;
End; {TDriveInfo.GetLongPrettyName}



Function TDriveInfo.GetData(Drive : TDrive) : TDriveInfoRec;
Begin
  if (Drive < 'A') Or (Drive > 'Z') then
    Raise EConvertError.Create(Format(ErrorInvalidDrive, [Drive]));

  Result := FData[Upcase(Drive)];
End; {TDriveInfo.GetData}


Function TDriveInfo.ReadDriveStatus(Drive : TDrive; Flags : Integer) : Boolean;
var  ErrorMode: word;
     shFlags  : Integer;
     FileInfo : TShFileInfo;
     MaximumComponentLength : DWORD;
     FileSystemNameBuffer : String;
     DriveID              : String;

Begin
  Drive := Upcase(Drive);
  DriveID := '';
  if (Drive < FirstDrive) Or (Drive > LastDrive) then
    Raise EConvertError.Create(Format(ErrorInvalidDrive, [Drive]));

  With FData[Drive] Do
  Begin
    Init      := True;
    DriveType := Windows.GetDriveType(PChar(Drive + ':\'));
    Valid     := DriveType in [DRIVE_REMOVABLE, DRIVE_FIXED, DRIVE_CDROM, DRIVE_RAMDISK, DRIVE_REMOTE];
    IF Valid Then
    Begin
      {Read driveStatus:}
      IF (Flags And dsSize) <> 0 Then
      Begin
        { turn off critical errors }
        ErrorMode := SetErrorMode(SEM_FailCriticalErrors or SEM_NOOPENFILEERRORBOX);
        try
         { drive 1 = a, 2 = b, 3 = c, etc. }
         Size       := DiskSize(Ord(Drive) - $40);
         DriveReady := Size > 0;
         IF DriveReady Then
         Begin
           SetLength(FilesystemNamebuffer,500) ;
           SetLength(DriveID, 24);
           {Access the physical drive:}
           if GetVolumeInformation(PChar(Drive + ':\'),
                                   PChar(DriveID), 24,
                                   @DriveSerial,
                                   maximumcomponentlength,
                                   FileSystemFlags,
                                   PChar(filesystemnamebuffer), 499) Then
           Begin
             FileSystemName := StrPas(PChar(FileSystemNameBuffer));
           End
           Else
           Begin
             DriveSerial := 0;
             FileSystemName := '';
           End;
           SetLength(FileSystemNameBuffer, 0);
         End
         Else
         Begin
           DriveSerial := 0;
         End;
        finally    { restore old error mode }
          SetErrorMode(ErrorMode);
        end;
      End;
      IF (Flags And dsDisplayName <> 0)Then
      Begin
        IF DriveReady or (Flags And dsSize = 0) Then
        Begin
         {Fetch drives displayname:}

         {Due to a bug in shGetFileInfo, this function returns allways the displayname of
          the first inserted disk, even if a disk change has occured. So, better use the
          Volume ID to build the drives displayname:}
         IF (DriveType = DRIVE_CDROM) And (Length(DriveID) > 0) Then
           DisplayName := DriveID[1] + LowerCase(Copy(DriveID, 2, 24)) + ' ('+ Drive + ':)'
         Else
         DisplayName := GetShellFileName(Drive + ':\');

         IF DriveType = DRIVE_REMOTE Then
           PrettyName  := Drive + ': ' + Copy(DisplayName, 1, Max(0, Min(Length(DisplayName), Pred(Pos(' ', DisplayName)))))
         Else
           PrettyName := Drive + ': ' + DisplayName;

         IF Pos(')', PrettyName) = Length(PrettyName) Then
         SetLength(PrettyName, Max(Length(PrettyName) - 5, 0));

         IF DriveType = DRIVE_REMOTE Then
           LongPrettyName := Copy(PrettyName, 1, 3) + GetNetWorkName(Drive)
         Else
         Begin
           LongPrettyName := Copy(PrettyName, 1, 3) + DisplayName;
           IF Pos(')', LongPrettyName) = Length(LongPrettyName) Then
           SetLength(LongPrettyName, Max(Length(LongPrettyName) - 5, 0));
         End;

        End
        Else
        Begin
         DisplayName := Drive + ':';
         PrettyName  := DisplayName;
         LongPrettyName := DisplayName;
        End;
      End;
      IF ((Flags And dsImageIndex) <> 0) And (ImageIndex = 0) Then
      Begin
        shFlags := SHGFI_SYSICONINDEX or SHGFI_SMALLICON;
        SHGetFileInfo(PChar(Drive + ':\'), 0, FileInfo, SizeOf(FileInfo), shFlags);
        ImageIndex := FileInfo.iIcon;
      End;
    End
    Else
    Begin
     Size        := 0;
     DriveReady  := False;
     DisplayName := '';
     PrettyName  := '';
     LongPrettyName := '';
     DriveSerial := 0;
    End;
    Result := Valid And DriveReady;
  End;
End; {TDriveInfo.ReadDriveStatus}

// ===========================================================
// Other service functions and procedures:
// ===========================================================





// returns the filename as displayed by the shell
function GetShellFileName ( const Name : string ):string;
var sfi : TSHFileInfo;
    E   : Integer;
begin
     E := SetErrorMode(SEM_FAILCRITICALERRORS);
     Try
       If SHGetFileInfo(PChar( Name ),0,sfi,SizeOf(TSHFileInfo),SHGFI_DISPLAYNAME) <> 0 Then
          Result := sfi.szDisplayName;
     finally
          SetErrorMode(E);
     End;
end; {GetShellFileName}



// Gets the network UNC-Name of a mounted drive:
Function GetNetWorkName(Drive : Char) : String;
Var P : Array[0..MAX_PATH] Of Char;
    MaxLen : DWORD;

Begin
  MaxLen := MAX_PATH;
  IF WNetGetConnection(PChar(String(Drive + ':')), P, MaxLen) = NO_ERROR Then
  Result := P
  Else
  Result := '';
End; {GetNetWorkName}


// ======================================================
// Initialization
// ======================================================

Initialization
  IF Not Assigned(DriveInfo) Then
  DriveInfo := TDriveInfo.Create;


// ======================================================
// Finalization
// ======================================================
Finalization
  IF Assigned(DriveInfo) Then
  Begin
    DriveInfo.Free;
    DriveInfo := NIL;
  End;


end.
